
// names.c
// Andrew Davison, ad@fivedots.coe.psu.ac.th, Feb 2020
/* A more complex data structure for displaying
   in gdb
*/

#include <stdio.h>
#include <stdlib.h>


struct person {
  char *name;
  int age;
  char *ssn;
};


struct node {
  struct person *person;
  struct node *next;
} *head, *tail;


void insert(struct person *p)
{
  struct node *node = malloc(sizeof(struct node));

  /* assign the person to the node */
  node->person = p;
  node->next = NULL;

  if (head == NULL) {   // add first elem
    head = node;
    tail = node;
  } 
  else {   // add subsequent elems
    tail->next = node;
    tail = node;
  }
}  // end of insert()


void displayNode(struct node *n)
{
  if (n == NULL)
    printf("Node is NULL\n");
  else
    printf("{%s,%d,%s}", n->person->name,
                         n->person->age,
                         n->person->ssn);
} // end of displayNode()



void displayAll(void)
{
  if (head == NULL) 
    printf("Empty list\n");
  else {
    struct node *curr = head;
    while (curr != NULL) {
      displayNode(curr);
      curr = curr->next;
      if (curr != NULL)
        printf("--");
    }
    printf("\n");
  }
}  // end of displayAll()



void display(int idx)
{
  if (head == NULL)
    printf("Empty list\n");
  else {
    struct node *curr = head;
    int i = 0;
    while ((i < idx) && (curr != NULL)) {
      i++;
      curr = curr->next;
    }
    if (curr == NULL)
      printf("No node at index %d\n", idx);
    else {
      displayNode(curr);
      printf("\n");
    }
  }
}  // end of display()



int main(void)
{
  int numPeople;
  printf("Enter no. of people: ");
  scanf("%d", &numPeople);

  for (int i = 0; i < numPeople; i++) {
    struct person *p = malloc(sizeof(struct person));
    p->name = malloc(100);
    p->ssn = malloc(100);

    printf("Person %d\n", (i+1));
    printf("  Enter name: ");
    scanf("%99s", p->name); 

    printf("  Enter age: ");
    scanf("%d", &p->age);

    printf("  Enter ssn: ");
    scanf("%99s", p->ssn); 

    insert(p);
  }

  displayAll();
  return 0;
}


